/**
  *  \file client/dlg-cargo.cc
  *  \brief Cargo transfer window.
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-cargo.h"
#include "client/cscreen.h"
#include "ui/widgets.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "ui/window.h"
#include "ui/layout.h"
#include "ui/event.h"
#include "ui/msgbox.h"
#include "cpluslib/format.h"
#include "ui/util.h"
#include "cpluslib/holder.h"
#include "ui/skin.h"
#include "cpluslib/smallset.h"
#include "ui/prefix.h"

namespace {
    enum {
        cm_Sup2mc         = 3000,
        cm_Unload         = 3001,
        cm_EnableOverload = 3002,

        cm_MoveLeft       = 2000,
        cm_MoveRight      = 2001
    };

    class WOverloadMarker : public UIFixedWidget {
        int counter;
        bool state;
     public:
        WOverloadMarker();
        void draw(GfxCanvas& can);
        bool handleEvent(const UIEvent& e, bool second_pass);
    };

    class WCargoTransferWindow : public UIWindow {
        GCargoTransfer& transfer;
        UIVBoxLayout    layout;
        WOverloadMarker overload_marker;
     public:
        WCargoTransferWindow(GCargoTransfer& transfer);
        ~WCargoTransferWindow();

        GCargoTransfer& getActor()
            { return transfer; }

        bool handleCommand(const UICommand& cmd, UIWidget& sender);

        bool isSup2mc() const;

        void enableOverload();

        UIWidget* sup2mc;
    };

    class WCargoLine : public UIFocusGroup {
        GCargoType       type;
        GCargoTransfer&  actor;
        UIButton         le, ri;
        void  showText(GfxCanvas& c, int side, int x, int y);
     public:
        WCargoLine(int id, GCargoType type, GCargoTransfer& actor);
        void  drawInterior(GfxCanvas& can);
        void  doLayout();
        void  getLayoutInfo(LayoutInfo& info);
        bool  handleEvent(const UIEvent& e, bool second_pass);
        bool  handleCommand(const UICommand& cmd, UIWidget& sender);
    };

    class WCargoHeader : public UIWidget {
        void  showTitle(GfxCanvas& can, GCargoContainer& con, int x);
     public:
        WCargoHeader();
        void  draw(GfxCanvas& can);
        void  getLayoutInfo(LayoutInfo& info);
    };
};

/** \class WCargoHeader
    \brief Header line (two green boxes)

    A WCargoHeader displays the header with the names of the cargo
    transfer partners, plus the column headings

    \inv owner must be a WCargoTransferWindow */
WCargoHeader::WCargoHeader()
    : UIWidget()
{
    setState(st_Disabled, true);
}

void
WCargoHeader::showTitle(GfxCanvas& can, GCargoContainer& con, int x)
{
    GfxParameters parm;
    GfxRect toprect(x-10, extent.y, 190, font_heights[FONT_NORMAL]*3-3);

    drawTiledArea(can, toprect, 0 /* pixmap, actually */,
                  standard_colors[COLOR_GREENSCALE+4], 0 /* alteration */);
    drawFrameDown(can, toprect);

    parm.color = standard_colors[COLOR_WHITE];
    parm.font = fonts[FONT_NORMAL];
    outTextF(can, parm, x, extent.y + 5, 180, con.getName());
    drawHLine(can, parm, x, extent.y + font_heights[FONT_NORMAL] + 4, x + 180);

    parm.just_x = 1;
    const int y = extent.y + font_heights[FONT_NORMAL] + 6;
    outText(can, parm, x+45, y, _("has"));
    outText(can, parm, x+135, y, _("space left"));
}

void
WCargoHeader::draw(GfxCanvas& can)
{
    WCargoTransferWindow& w = *checked_cast<WCargoTransferWindow*>(getOwner());
    showTitle(can, w.getActor().getContainer(0), extent.x + 10);
    showTitle(can, w.getActor().getContainer(1), extent.x + 430);
    can.update(getExtent());
}

void
WCargoHeader::getLayoutInfo(LayoutInfo& info)
{
    // FIXME: width is same as WCargoLine
    info.flags = li_FlexibleH;
    info.min_size = info.pref_size = GfxPoint(430+190, 3*font_heights[FONT_NORMAL]);
}

/******************************* WCargoLine ******************************/

/** \class WCargoLine
    \brief One line in the cargo window

    Displays amounts for one cargo type, as well as two buttons for
    manipulating it. */

/** Constructor.
    \param id      widget id
    \param type    type of cargo
    \param actor   underlying transaction class */
WCargoLine::WCargoLine(int id, GCargoType type, GCargoTransfer& actor)
    : UIFocusGroup(id), type(type), actor(actor),
      le(cm_MoveLeft,  "\033", UIKey(SDLK_LEFT)),
      ri(cm_MoveRight, "\032", UIKey(SDLK_RIGHT))
{
    le.addCommand();
    ri.addCommand();
    add(le);
    add(ri);
}

void
WCargoLine::drawInterior(GfxCanvas& can)
{
    UISkin& s = getSkin();

    s.startDraw();
    s.setColor(UIColor::tc_Static);
    s.setSolid(false);

    string_t unit = getCargoUnit(type);
    if (unit.length())
        unit = " [" + unit + "]";

    s.outTextF(can, FONT_NORMAL, extent.x + 230, extent.y + 2, 170,
               getCargoName(type) + unit);

    s.setSolid();
    showText(can, 0, extent.x + 10,  extent.y + 2);
    showText(can, 1, extent.x + 430, extent.y + 2);

    UIFocusGroup::drawInterior(can); /* calls update */
}

void
WCargoLine::showText(GfxCanvas& can, int side, int x, int y)
{
    int32 now = actor.getContainer(side).getCargo(type);
    int32 max = actor.getContainer(side).getMaxCargo(type);

    UISkin& s = getSkin();
    s.setColor(UIColor::tc_Green);
    s.getParameters().just_x = 2;
    s.getParameters().just_y = 0;
    s.outTextF(can, FONT_NORMAL, x + 90, y, 90, itoa(now));

    if (max > 20000)
        s.outTextF(can, FONT_NORMAL, x + 180, y, 90, _("(unl)"));
    else
        s.outTextF(can, FONT_NORMAL, x + 180, y, 90, itoa(max - now));
}

/* Display:

   Item              PCC1                     PCC2

   left              +19                      +10
                       +0  W90   amount        +0 W90
                       +90 W90   maximum       +90 W90
   button            +210  W19                +200 W20
   good name/unit    +240                     +230
   button            +410  W19                +400 W20
   right             +440                     +430          */
void
WCargoLine::doLayout()
{
    le.setExtent(GfxRect(extent.x + 200, extent.y + 2, 20, 16), true);
    ri.setExtent(GfxRect(extent.x + 400, extent.y + 2, 20, 16), true);
    drawWidget(false);
}

void
WCargoLine::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(430+190, 20);
}

bool
WCargoLine::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && hasState(st_Focused)) {
        WCargoTransferWindow& w = *checked_cast<WCargoTransferWindow*>(getOwner());
        bool s2m = w.isSup2mc();

        switch(e.key.code) {
         case SDLK_LEFT:
            handleCommand(UICommand(cm_MoveLeft, consumePrefixArg()), *this);
            return true;
         case SDLK_RIGHT:
            handleCommand(UICommand(cm_MoveRight, consumePrefixArg()), *this);
            return true;
         case ss_Shift + SDLK_LEFT:
            actor.move(type, 1, 1, true, s2m);
            return true;
         case ss_Shift + SDLK_RIGHT:
            actor.move(type, 1, 0, true, s2m);
            return true;
         case ss_Ctrl + SDLK_LEFT:
            actor.move(type, 100, 1, true, s2m);
            return true;
         case ss_Ctrl + SDLK_RIGHT:
            actor.move(type, 100, 0, true, s2m);
            return true;
         case ss_Alt + SDLK_LEFT:
            if (int n = consumePrefixArg())
                actor.move(type, n - actor.getContainer(0).getCargo(type), 1, true, s2m);
            else
                actor.move(type, actor.getContainer(1).getCargo(type), 1, true, s2m);
            return true;
         case ss_Alt + SDLK_RIGHT:
            if (int n = consumePrefixArg())
                actor.move(type, n - actor.getContainer(1).getCargo(type), 0, true, s2m);
            else
                actor.move(type, actor.getContainer(0).getCargo(type), 0, true, s2m);
            return true;
         case ss_Alt + 'o':
            handleCommand(UICommand(cm_EnableOverload), *this);
            return true;
        }
    }
    return UIFocusGroup::handleEvent(e, second_pass);
}

bool
WCargoLine::handleCommand(const UICommand& cmd, UIWidget& sender)
{
    if (cmd.code == cm_MoveLeft || cmd.code == cm_MoveRight) {
        WCargoTransferWindow& w = *checked_cast<WCargoTransferWindow*>(getOwner());
        bool s2m = w.isSup2mc();
        int dir = (cmd.code == cm_MoveLeft);
        int amount = cmd.arg ? cmd.arg : 10;
        actor.move(type, amount, dir, true, s2m);
        return true;
    } else {
        return UIFocusGroup::handleCommand(cmd, sender);
    }
}

/**************************** WOverloadMarker ****************************/

WOverloadMarker::WOverloadMarker()
    : counter(0), state(true)
{ }

void
WOverloadMarker::draw(GfxCanvas& can)
{
    GfxParameters parm;
    parm.color = state ? standard_colors[COLOR_WHITE] : standard_colors[COLOR_BLACK];
    parm.font  = fonts[FONT_TITLE];
    outTextF(can, parm, extent.x, extent.y, extent.w, _("Overload Mode"));
    can.update(extent);
}

bool
WOverloadMarker::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Tick) {
        if (++counter > 40) {
            counter = 0;
            state = !state;
            drawWidget(false);
        }
    }
    return UIFixedWidget::handleEvent(e, second_pass);
}

/************************** WCargoTransferWindow *************************/

/** \class WCargoTransferWindow
    \brief Cargo Transfer Window (normal transfer) */

/** Constructor.
    \param left,right cargo transfer traits */
WCargoTransferWindow::WCargoTransferWindow(GCargoTransfer& transfer)
    : UIWindow(_("Cargo Transfer"), BLUE_WINDOW, COLOR_GRAY),
      transfer(transfer),
      layout(0, 5),
      sup2mc(0)
{
    setLayout(layout);
}

WCargoTransferWindow::~WCargoTransferWindow()
{ }

bool
WCargoTransferWindow::handleCommand(const UICommand& cmd, UIWidget& sender)
{
    if (cmd.code == cm_Sup2mc && sup2mc != 0) {
        sup2mc->setState(UIButton::bf_Highlight, !sup2mc->hasState(UIButton::bf_Highlight));
        return true;
    }
    if (cmd.code == cm_Unload) {
        transfer.doUnload(isSup2mc());
        return true;
    }
    if (cmd.code == cm_EnableOverload) {
        enableOverload();
        return true;
    }
    return UIWindow::handleCommand(cmd, sender);
}

bool
WCargoTransferWindow::isSup2mc() const
{
    return sup2mc && sup2mc->hasState(UIButton::bf_Highlight);
}

void
WCargoTransferWindow::enableOverload()
{
    /* already enabled? */
    if (transfer.getContainer(0).hasFlag(GCargoContainer::can_Overload))
        return;

    /* ask user */
    if (!UIMessageBox(_("Overload Mode allows you to load more cargo on ships "
                       "than PCC usually permits. There are a few cases "
                       "where this is allowed and useful, otherwise the Host "
                       "will detect this and destroy your cargo.\n"
                       "Turn on Overload Mode?"),
                      _("Cargo Transfer")).doYesNoDialog())
        return;

    /* ok, enable it */
    for (int i = 0; i < 2; ++i) {
        GCargoContainer& c = transfer.getContainer(i);
        c.setFlags(c.getFlags() | GCargoContainer::can_Overload);
    }
    getActor().sig_changed.raise();

    overload_marker.setExtent(GfxRect(extent.x + 5, extent.y + 2,
                                      extent.w/2, 24 /* FIXME hardcoded values */));
    add(overload_marker);
}

/** The order in which cargo types appear on the window. This is the
    same order as used in PCC 1.x. */
static const GCargoType elements[] = {
    el_Neutronium,  el_Tritanium,  el_Duranium,  el_Molybdenum,
    el_Supplies,    el_Colonists,  el_Money,
    el_Fighters,
    el_Torps,
    GCargoType(el_Torps+1),
    GCargoType(el_Torps+2),
    GCargoType(el_Torps+3),
    GCargoType(el_Torps+4),
    GCargoType(el_Torps+5),
    GCargoType(el_Torps+6),
    GCargoType(el_Torps+7),
    GCargoType(el_Torps+8),
    GCargoType(el_Torps+9)
};

/** Normal cargo transfer, main entry point. Does all of the cargo
    transfer actions between left and right. By convention, left is
    the object the user called the cargo transfer from. */
void
doCargoTransfer(GTransaction& main_ta, GCargoTransfer& transfer)
{
    GCargoContainer& left  = transfer.getContainer(0);
    GCargoContainer& right = transfer.getContainer(1);

    /* figure out what can be transferred */
    GCargoTypeSet available_types;
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i)
        if (left.canHaveCargo(i) && right.canHaveCargo(i) && (left.getCargo(i) || right.getCargo(i)))
            available_types += i;
    if (available_types.contains(el_Supplies)
        && ((left.getFlags() | right.getFlags()) & GCargoContainer::can_Sup2mc))
        available_types += el_Money;

    if (available_types.empty()) {
        messageBox(format(_("There is nothing you could transfer to or from %s."),
                          right.getName()),
                   _("Cargo Transfer"));
        return;
    }

    /* create window */
    ObjectHolder holder;
    WCargoTransferWindow win(transfer);
    win.add(holder.add(new WCargoHeader()));
    win.add(holder.add(new UIPrefixArg()));
    int line_id = 1000;
    for (unsigned i = 0; i < countof(elements); ++i) {
        // WARNING: WCargoHeader and WCargoLine both assume that their immediate
        // parent is WCargoTransferWindow.
        if (available_types.contains(elements[i])) {
            UIWidget& l = holder.add(new WCargoLine(line_id++, elements[i], win.getActor()));
            win.getActor().sig_changed.add(&l, (void (UIWidget::*)()) &UIWidget::drawWidget);
            win.add(l);
        }
    }
    win.add(holder.add(new UIFocusIterator(1000, line_id-1, UIFocusIterator::fi_Vert)));

    UIGroup button_list(UIHBoxLayout::instance5);

    // FIXME: this layout is rather minimalistic and does not match PCC 1.x
    // PCC 1.x has
    //     [Sup2mc]      [OK] [Cancel] [H]    [<-Unload]
    // or  [Unload->]    [OK] [Cancel] [H]      [Sup2mc]
    // where both Sup2mc and Unload are optional, and the buttons are aligned
    // with the WCargoLine controls. This is hard to express using standard
    // layout managers, we probably need one specially-made for this situation.
    if (left.hasFlag(GCargoContainer::can_Sup2mc) || right.hasFlag(GCargoContainer::can_Sup2mc)) {
        win.sup2mc = &holder.add(new UIButton(cm_Sup2mc, _("Sup->mc"), 's')).addCommand();
    }

    if (left.hasFlag(GCargoContainer::can_Unload) && right.hasFlag(GCargoContainer::can_GetUnload)) {
        /* [Sup2mc] [Unload->] [....] [OK] [Cancel] */
        if (win.sup2mc)
            button_list.add(*win.sup2mc);
        button_list.add(holder.add(new UIButton(cm_Unload, _("Unload->"), 'u')).addCommand());
    } else if (right.hasFlag(GCargoContainer::can_Unload) && left.hasFlag(GCargoContainer::can_GetUnload)) {
        /* [Unload<-] [Sup2mc] [....] [OK] [Cancel] */
        button_list.add(holder.add(new UIButton(cm_Unload, _("<-Unload"), 'u')).addCommand());
        if (win.sup2mc)
            button_list.add(*win.sup2mc);
    } else {
        /* [Sup2mc] [....] [OK] [Cancel] */
        if (win.sup2mc)
            button_list.add(*win.sup2mc);
    }
    button_list.add(holder.add(new UISpacer()));
    button_list.add(holder.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    button_list.add(holder.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    win.add(button_list);

    /* run it */
    win.pack();
    ui_root->moveWidgetToEdge(win, 1, 0, 0);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    if (rv)
        main_ta.doCommit();
    else
        main_ta.doCancel();
}
