/**
  *  \file client/wracelist.cc
  *  \brief Color-keyed race list widget
  */

#include "client/wracelist.h"
#include "game/player.h"
#include "gfx/gfxutil.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"

/** Color mapping. Indexed by player-number-minus-one. In PCC 1.x, the statistic
    and combat diagram screens have a separate palette for races 7-11. Since we
    usually have close colors, we use those instead. */
const uint8 WRaceList::colors[] = {
    COLOR_BLACK,          /* Fed: black */
    COLOR_WHITE,          /* Lizard: white */
    COLOR_YELLOW,         /* Bird: yellow */
    COLOR_GREEN,          /* Klingon: green */
    COLOR_BLUE,           /* Pirate: blue */
    COLOR_RED,            /* Cyborg: red */
    COLOR_BRIGHTMAGENTA,  /* Crystal: lilac. 63/24/63 [255/97/255] in PCC 1.x, 255/85/255 here */
    COLOR_BRIGHTBROWN,    /* Empire: brown. 48/24/0 [194/97/255] in PCC 1.x, 194/97/255 here */
    COLOR_BRIGHTORANGE,   /* Robot: orange. 63/48/0 [255/194/0] in PCC 1.x, 255/194/0 here */
    COLOR_BRIGHTCYAN,     /* Rebel: cyan. 32/63/63 [128/255/255] in PCC 1.x, 85/255/255 here */
    COLOR_DARKGREEN,      /* Colony: green. 0/48/0 [0/194/0] in PCC 1.x, 0/170/0 here */
    COLOR_BLUEBLACK       /* Aliens: blue/gray */
};

/** Constructor.
    \param preferred_width preferred width of widget, in pixels
    \param from first player to display, [1,NUM_OWNERS]
    \param to last player to display, inclusive, [from,NUM_OWNERS] */
WRaceList::WRaceList(int preferred_width, int from, int to)
    : preferred_width(preferred_width),
      from(from),
      to(to)
{ }

/** Compute layout of widget.
    \param positions [out] Relative positions of player boxes
    \param w [in] Assumed widget width in pixels */
void
WRaceList::calcLayout(GPlayerArray<GfxRect>& positions, int w)
{
    int x = 0, y = 0;
    int cellheight = font_heights[FONT_NORMAL];

    for (int i = from; i <= to; ++i) {
        int cellwidth = fonts[FONT_NORMAL]->getTextWidth(getAdjName(i));
        if (cellwidth > 0) {
            /* cell with content */
            if (x + cellwidth > w) {
                /* start new line */
                x = 0;
                y += cellheight;
            }
            positions[i] = GfxRect(x, y, cellwidth, cellheight);
            x += cellwidth;
            x += font_ems[FONT_NORMAL]/2;
        } else {
            /* empty cell */
            positions[i] = GfxRect(x, y, 0, cellheight);
        }
    }
}

void
WRaceList::draw(GfxCanvas& can)
{
    GfxParameters parm;
    parm.font = fonts[FONT_NORMAL];
    calcLayout(positions, extent.w);
    for (int i = from; i <= to; ++i) {
        GfxPoint pos = positions[i].getTopLeft().movedBy(extent.getTopLeft());
        parm.color = standard_colors[colors[i-1]];
        outText(can, parm, pos.x, pos.y, getAdjName(i));
    }
    can.update(extent);
}

void
WRaceList::getLayoutInfo(LayoutInfo& info)
{
    GPlayerArray<GfxRect> pos;
    calcLayout(pos, preferred_width);

    info.pref_size = info.min_size = GfxPoint(preferred_width,
                                              pos[to].getBottomY());
    info.flags = 0;
}

bool
WRaceList::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse) {
        if (extent.contains(e.mouse.x, e.mouse.y)) {
            if (e.mouse.b != 0) {
                /* mouse clicked into this widget */
                int where = 0;
                GfxPoint rel(e.mouse.x - extent.x,
                             e.mouse.y - extent.y);
                for (int i = from; i <= to; ++i) {
                    if (positions[i].contains(rel)) {
                        where = i;
                        break;
                    }
                }
                if (where != current_click && where != 0)
                    sig_click.raise(where);
                current_click = where;
                return true;
            } else {
                current_click = 0;
                return true;
            }
        } else {
            current_click = 0;
            return UIWidget::handleEvent(e, second_pass);
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}
