/**
  *  \file int/stmt.cc
  *  \brief Compilation of Statements into Bytecode
  *
  *  This compiles statements into bytecode.
  *  - bytecode.h defines the bytecode
  *  - exec.h runs it
  *  - expr.h compiles expressions.
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/stmt.h"
#include "int/bytecode.h"
#include "int/lex.h"
#include "cpluslib/string.h"
#include "int/expr.h"
#include "int/csource.h"
#include "util/bsearch.h"
#include "int/exec.h"

enum IntKeyword {
    kw_Abort,
    kw_Bind,
    kw_Break,
    kw_Close,
    kw_Continue,
    kw_CreateShipProperty,
    kw_CreatePlanetProperty,
    kw_CreateKeymap,
    kw_Dim,
    kw_Do,
    kw_Else,
    kw_End,
    kw_EndIf,
    kw_EndSub,
    kw_EndTry,
    kw_EndWith,
    kw_Eval,
    kw_For,
    kw_ForEach,
    kw_Function,
    kw_Get,
    kw_If,
    kw_Input,
    kw_Load,
    kw_Local,
    kw_Loop,
    kw_Next,
    kw_On,
    kw_Open,
    kw_Option,
    kw_Print,
    kw_Put,
    kw_Return,
    kw_RunHook,
    kw_Seek,
    kw_SetStr,
    kw_SetByte,
    kw_SetWord,
    kw_SetLong,
    kw_Shared,
    kw_Static,
    kw_Stop,
    kw_Sub,
    kw_Try,
    kw_TryLoad,
    kw_Until,
    kw_While,
    kw_With,
    kw_NO_KEYWORD
};

static const char*const keyword_table[] = {
    "ABORT",
    "BIND",
    "BREAK",
    "CLOSE",
    "CONTINUE",
    "CREATESHIPPROPERTY",
    "CREATEPLANETPROPERTY",
    "CREATEKEYMAP",
    "DIM",
    "DO",
    "ELSE",
    "END",
    "ENDIF",
    "ENDSUB",
    "ENDTRY",
    "ENDWITH",
    "EVAL",
    "FOR",
    "FOREACH",
    "FUNCTION",
    "GET",
    "IF",
    "INPUT",
    "LOAD",
    "LOCAL",
    "LOOP",
    "NEXT",
    "ON",
    "OPEN",
    "OPTION",
    "PRINT",
    "PUT",
    "RETURN",
    "RUNHOOK",
    "SEEK",
    "SETSTR",
    "SETBYTE",
    "SETWORD",
    "SETLONG",
    "SHARED",
    "STATIC",
    "STOP",
    "SUB",
    "TRY",
    "TRYLOAD",
    "UNTIL",
    "WHILE",
    "WITH"
};
CTASSERT(countof(keyword_table) == kw_NO_KEYWORD);

static IntKeyword
lookupKeyword(string_t name)
{
    int n = bin_search(&keyword_table[0], countof(keyword_table),
                       IntNameLookupComparator(name));
    if (n < 0)
        return kw_NO_KEYWORD;
    else
        return IntKeyword(n);
}

/************************** IntStatementCompiler *************************/

/** Create IntStatementCompiler.
    \param source   where to get commands from
    \param bco      output BCO
    \param context_level number of contexts which must be removed for `Return'
    \param break_addr    address for `Break'
    \param continue_addr address for `Continue'
    \param with_level    number of active contexts which must be removed for
                         Break or Continue. */
IntStatementCompiler::IntStatementCompiler(IntCommandSource& source,
                                           IntByteCodeObject& bco,
                                           int context_level,
                                           IntByteCodeObject::label_t break_addr,
                                           IntByteCodeObject::label_t continue_addr,
                                           int with_level)
    : lex(string_t()), source(source), bco(bco),
      break_addr(break_addr),
      continue_addr(continue_addr),
      with_level(with_level),
      context_level(context_level)
{ }

IntStatementCompiler::~IntStatementCompiler()
{ }

void
IntStatementCompiler::nextLine()
{
    string_t s;
    if (!source.getLine(s))
        compileError("Unexpected end of source code");
    lex.setLine(s);
    lex.lex();
}

bool
IntStatementCompiler::parseStatementsUntil(const char* term1, const char* term2)
{
    while(1) {
        nextLine();
        if (term1 && lex.lookingAt(term1))
            return false;
        if (term2 && lex.lookingAt(term2))
            return true;
        parseStatement(true);
    }
}

void
IntStatementCompiler::parseExpression()
{
    IntExprParser xp(lex);
    xp.parseExpr();
    xp.pop()->compileV(bco);
}

/** Parse body and finish of a "Do" loop. This must be executed in the
    context of a different IntStatementCompiler than the surrounding
    statement, for correct "break" and "continue" handling. */
void
IntStatementCompiler::parseDoBody(IntByteCodeObject::label_t again)
{
    parseStatementsUntil("LOOP", 0);

    /* trailing condition */
    if (lex.getType() == tt_Null) {
        bco.setLabel(continue_addr);
        bco.add(mkCode(op_Jump, again));
        bco.setLabel(break_addr);
    } else if (lex.lookingAt("WHILE")) {
        bco.setLabel(continue_addr);
        parseExpression();
        bco.add(mkCode(op_JumpTPop, again));
        if (lex.getType() != tt_Null)
            compileError("expression incorrectly terminated");
        bco.setLabel(break_addr);
    } else if (lex.lookingAt("UNTIL")) {
        bco.setLabel(continue_addr);
        parseExpression();
        bco.add(mkCode(op_JumpFEPop, again));
        if (lex.getType() != tt_Null)
            compileError("expression incorrectly terminated");
        bco.setLabel(break_addr);
    } else {
        compileError("expected \"While\" or \"Until\" after \"Loop\"");
    }
}

void
IntStatementCompiler::parseWith(bool oneline, string_t content)
{
    bco.add(mkCode(op_PushContext, 0));
    if (oneline) {
        lex.setLine(content);
        lex.lex();
        parseStatement(false);
    } else {
        parseStatementsUntil("ENDWITH", 0);
    }
    bco.add(mkCode(op_PopContext, 1));
}

void
IntStatementCompiler::parseForEach(bool oneline, string_t line)
{
    IntByteCodeObject::label_t over = bco.allocLabel(), again = bco.allocLabel();
    break_addr = bco.allocLabel();
    continue_addr = bco.allocLabel();

    /* when we're here, TOS contains the array to loop through */
    bco.add(mkCode(op_FirstContext, 0));
    bco.add(mkCode(op_JumpFEPop, over));
    if (oneline) {
        bco.setLabel(again);
        lex.setLine(line);
        lex.lex();
        parseStatement(false);
    } else {
        parseStatementsUntil("NEXT", 0);
    }
    bco.setLabel(continue_addr);
    bco.add(mkCode(op_NextContext, 0));
    bco.add(mkCode(op_JumpTPop, again));
    bco.setLabel(break_addr);
    bco.add(mkCode(op_PopContext, 1));
    bco.setLabel(over);
}

IntByteCodeObject&
IntStatementCompiler::parseSub()
{
    // FIXME: this will leave an inaccessible BCO around which will
    // (correctly) have ref-count zero. We should add BCO GC someday.
    IntByteCodeObject& bco = IntByteCodeObject::allocBCO();
    if (lex.getType() == tt_Left) {
        /* parse parameter names. */
        bool optional = false;
        lex.lex();
        while (1) {
            if (lex.lookingAt("OPTIONAL")) {
                if (optional)
                    compileError("`Optional' already seen");
                optional = true;
            }
            if (lex.getType() != tt_Id)
                compileError("expected parameter name");
            if (lookupKeyword(lex.getString()) != kw_NO_KEYWORD)
                compileError("keyword may not be used as parameter name");
            for (uint32 i = 0; i < bco.getNumArgs(); ++i)
                if (bco.getArgName(i) == lex.getString())
                    compileError("duplicate parameter name");
            bco.addArgument(lex.getString(), !optional);
            lex.lex();
            if (lex.lookingAt(tt_Right))
                break;
            if (!lex.lookingAt(tt_Comma))
                compileError("expected `,' or ')'");
        }
        if (lex.getType() != tt_Null)
            compileError("expected end-of-line after subroutine");
    } else if (lex.getType() == tt_Null) {
        /* empty */
    } else {
        compileError("expected `(' after `Sub'");
    }

    /* the header has been parsed, parse the body */
    IntStatementCompiler(source, bco, 0).parseStatementsUntil("ENDSUB", 0);
    return bco;
}

void
IntStatementCompiler::parseDim(IntOpcode op)
{
    while (1) {
        if (lex.getType() != tt_Id)
            compileError("Expected variable name");
        string_t name = lex.getString();
        if (lookupKeyword(name) != kw_NO_KEYWORD)
            compileError("Keyword not allowed as variable name");
        lex.lex();
        if (lex.getType() == tt_EQ || lex.getType() == tt_Asgn) {
            lex.lex();
            parseExpression();
        } else {
            bco.add(mkCode(op_PushNull, 0));
        }
        bco.add(mkCode(op, bco.allocStr(name)));
        if (lex.getType() == tt_Null)
            break;
        if (lex.getType() != tt_Comma)
            compileError("expected `,'");
        lex.lex();
    }
}

void
IntStatementCompiler::parseStatement(bool allow_multiline)
{
    if (lex.getType() == tt_Null) {
        /* null statement */
        return;
    }
    if (allow_multiline)
        bco.add(mkCode(op_SetLine, source.getLineNr()));
    if (lex.getType() == tt_Id) {
        /* statement or expression */
        switch (lookupKeyword(lex.getString())) {
         case kw_If: {
            lex.lex();
            parseExpression();
            if (lex.lookingAt("THEN") && lex.getType() != tt_Null) {
                /* one-line IF */
                IntByteCodeObject::label_t end = bco.allocLabel();
                bco.add(mkCode(op_JumpFEPop, end));
                parseStatement(false);
                bco.setLabel(end);
            } else if (lex.getType() == tt_Null) {
                /* multi-line IF */
                if (!allow_multiline)
                    compileError("Multi-line \"If\" not allowed here");
                IntByteCodeObject::label_t not_t = bco.allocLabel();
                bco.add(mkCode(op_JumpFEPop, not_t));
                if (parseStatementsUntil("ENDIF", "ELSE")) {
                    /* else */
                    if (lex.getType() != tt_Null)
                        compileError("Expected end-of-line after \"Else\"");
                    IntByteCodeObject::label_t end = bco.allocLabel();
                    bco.add(mkCode(op_Jump, end));
                    bco.setLabel(not_t);
                    parseStatementsUntil("ENDIF", 0);
                    bco.setLabel(end);
                } else {
                    bco.setLabel(not_t);
                }
                if (lex.getType() != tt_Null)
                    compileError("Expected end-of-line after \"EndIf\"");
            } else {
                compileError("Expected \"Then\" or end of line");
            }
            break;
         }
         case kw_Dim: {
            lex.lex();
            if (lex.lookingAt("LOCAL"))
                parseDim(op_DimLocal);
            else if (lex.lookingAt("SHARED"))
                parseDim(op_DimShared);
            else if (lex.lookingAt("STATIC"))
                parseDim(op_DimStatic);
            else
                parseDim(op_DimLocal);
            break;
         }
         case kw_Do: {
            if (!allow_multiline)
                compileError("Multi-line \"Do\" not allowed here");
            IntByteCodeObject::label_t ba = bco.allocLabel(), ca = bco.allocLabel(),
                again = bco.allocLabel();
            /* leading condition */
            lex.lex();
            if (lex.getType() == tt_Null) {
                bco.setLabel(again);
            } else if (lex.lookingAt("WHILE")) {
                bco.setLabel(again);
                parseExpression();
                bco.add(mkCode(op_JumpFEPop, ba));
                if (lex.getType() != tt_Null)
                    compileError("expression incorrectly terminated");
            } else if (lex.lookingAt("UNTIL")) {
                bco.setLabel(again);
                parseExpression();
                bco.add(mkCode(op_JumpTPop, ba));
                if (lex.getType() != tt_Null)
                    compileError("expression incorrectly terminated");
            } else {
                compileError("expected \"While\" or \"Until\" after \"Do\"");
            }

            /* content */
            IntStatementCompiler(source, bco, context_level, ba, ca).parseDoBody(again);
            break;
         }
         case kw_ForEach: {
            lex.lex();
            parseExpression();
            string_t rem = lex.getRemainder();
            if (lex.lookingAt("DO") && lex.getType() != tt_Null) {
                /* one-line loop */
                IntStatementCompiler(source, bco, context_level+1).parseForEach(true, rem);
            } else if (lex.getType() == tt_Null) {
                /* multi-line loop */
                if (!allow_multiline)
                    compileError("Multi-line `ForEach' not allowed here");
                IntStatementCompiler(source, bco, context_level+1).parseForEach(false, string_t());
            } else {
                compileError("Expected `Do' after `ForEach'");
            }
            break;
         }
         case kw_Local: {
            lex.lex();
            parseDim(op_DimLocal);
            break;
         }
         case kw_Shared: {
            lex.lex();
            parseDim(op_DimShared);
            break;
         }
         case kw_Static: {
            lex.lex();
            parseDim(op_DimStatic);
            break;
         }
         case kw_Sub: {
            if (!allow_multiline)
                compileError("`Sub' not allowed here");
            if (lex.lex() != tt_Id)
                compileError("expected identifier after `Sub'");
            if (lookupKeyword(lex.getString()) != kw_NO_KEYWORD)
                compileError("keyword may not be used as subroutine name");
            bco.add(mkCode(op_PushString, bco.allocStr(lex.getString())));
            lex.lex();
            IntByteCodeObject& sub_bco = parseSub();
            bco.add(mkCode(op_DefineSub, sub_bco.getId()));
            break;
         }
         case kw_Return: {
            // FIXME: determine whether we're a function
            if (lex.lex() != tt_Null)
                compileError("expected end-of-line");
            if (context_level)
                bco.add(mkCode(op_PopContext, context_level));
            bco.add(mkCode(op_Return, 0));
            break;
         }
         case kw_Try: {
            if (lex.lex() != tt_Null) {
                /* one-line try */
                IntByteCodeObject::label_t end = bco.allocLabel();
                bco.add(mkCode(op_Try, end));
                parseStatement(false);
                bco.add(mkCode(op_EndTry, 0));
                bco.add(mkCode(op_PushNull, 0));
                bco.setLabel(end);
                bco.add(mkCode(op_Pop, 1));
            } else {
                /* multi-line try */
                if (!allow_multiline)
                    compileError("multi-line `Try' not allowed here");
                IntByteCodeObject::label_t end = bco.allocLabel();
                bco.add(mkCode(op_Try, end));
                if (parseStatementsUntil("ENDTRY", "ELSE")) {
                    /* it has an "Else" branch */
                    if (lex.lex() != tt_Null)
                        compileError("expected end-of-line after \"Else\"");
                    IntByteCodeObject::label_t real_end = bco.allocLabel();
                    bco.add(mkCode(op_EndTry, 0));
                    bco.add(mkCode(op_Jump, real_end));
                    bco.setLabel(end);
                    bco.add(mkCode(op_UnqualAssign, bco.allocStr("SYSTEM.ERR")));
                    parseStatementsUntil("ENDTRY", 0);
                    bco.setLabel(real_end);
                } else {
                    bco.add(mkCode(op_EndTry, 0));
                    bco.add(mkCode(op_PushNull, 0));
                    bco.setLabel(end);
                    bco.add(mkCode(op_Pop, 1));
                }
            }
            break;
         }
         case kw_With: {
            lex.lex();
            parseExpression();
            string_t rem = lex.getRemainder();
            if (lex.lookingAt("DO") && lex.getType() != tt_Null) {
                /* one-line */
                IntStatementCompiler(source, bco, context_level+1, break_addr, continue_addr, with_level+1).parseWith(true, rem);
            } else if (lex.getType() == tt_Null) {
                /* multi-line */
                if (!allow_multiline)
                    compileError("Multi-line `With' not allowed here");
                IntStatementCompiler(source, bco, context_level+1, break_addr, continue_addr, with_level+1).parseWith(false, string_t());
            } else {
                compileError("Expected `Do' after `With'");
            }
            break;
         }
         case kw_Break: {
            if (break_addr == IntByteCodeObject::NO_LABEL)
                compileError("`Break' not allowed here");
            if (lex.lex() != tt_Null)
                compileError("`Break' does not take arguments");
            if (with_level)
                bco.add(mkCode(op_PopContext, with_level));
            bco.add(mkCode(op_Jump, break_addr));
            break;
         }
         case kw_Continue: {
            if (break_addr == IntByteCodeObject::NO_LABEL)
                compileError("`Continue' not allowed here");
            if (lex.lex() != tt_Null)
                compileError("`Continue' does not take arguments");
            if (with_level)
                bco.add(mkCode(op_PopContext, with_level));
            bco.add(mkCode(op_Jump, continue_addr));
            break;
         }
         case kw_Stop: {
            if (lex.lex() != tt_Null)
                compileError("`Stop' does not take arguments");
            bco.add(mkCode(op_SetStatus, IntExecutionContext::ps_Stopped));
            break;
         }
         case kw_End: {
            if (lex.lex() != tt_Null)
                compileError("`End' does not take arguments");
            bco.add(mkCode(op_SetStatus, IntExecutionContext::ps_Finished));
            break;
         }
         case kw_Abort: {
            lex.lex();
            parseExpression();
            if (lex.getType() != tt_Null)
                compileError("expected end-of-line");
            bco.add(mkCode(op_Throw, 0));
            break;
         }
         /* The following keywords are handled by their containing
            syntactic elements, i.e. 'else' is handled by 'Try' or
            'If', and so on. */
         case kw_Else:
         case kw_EndIf:
         case kw_EndSub:
         case kw_EndTry:
         case kw_EndWith:
         case kw_Loop:
         case kw_Next:
         case kw_Until:
         case kw_While:
            compileError("Misplaced reserved word");
            break;
         case kw_NO_KEYWORD:
            /* don't know what it is.
               FIXME: optimize common case `i := expr' */
            string_t line = lex.getString() + lex.getRemainder();
            bco.add(mkCode(op_EvalStr, bco.allocStr(line)));
            break;
        }
        return;
    }

    /* must be an expression */
    IntExprParser xp(lex);
    xp.parseExpr();
    xp.pop()->compileS(bco);
}

bool
IntStatementCompiler::parseWholeStatement(bool allow_multiline)
{
    string_t str;
    if (!source.getLine(str))
        return false;
    lex.setLine(str);
    lex.lex();
    parseStatement(allow_multiline && source.isMultiline());
    return true;
}

void
IntStatementCompiler::compileError(const char* text)
{
    compileError(string_t(text));
}

void
IntStatementCompiler::compileError(const string_t& s)
{
    throw IntException(s);
}
