/**
 *  \file testaa.cc
 *  \brief Bresenham AA line drawing
 *  \author Stefan Reuther
 */

#include <algorithm>                  // std::swap
#include <math.h>                     // sin, cos for test
#include "gfx/init.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"


/** Line drawing, with anti-aliasing. Draws a line from x1,y1 to
    x2,y2, inclusive. For simplicity, this maps all lines into four
    octants, and decides between two cases handling two of them each. 
    To allow for lines that nicely connect to each other, we'd need to
    exclude x2,y2 from drawing, and implement eight octants (as four
    cases). The initialisation of the error term might also need
    adjustment. */
static void
drawLineAA(GfxCanvas& can, GfxParameters& parm, int x1, int y1, int x2, int y2)
{
    /* Handle easy cases first */
    if (x1 == x2) {
        drawVLine(can, parm, x1, y1, y2);
        return;
    }
    if (y1 == y2) {
        drawHLine(can, parm, x1, y1, x2);
        return;
    }

    /* Vanilla Bresenham */
    int dx = x2 - x1;
    int dy = y2 - y1;
    if(dx < 0)
        dx = -dx;
    if(dy < 0)
        dy = -dy;

    if(dx < dy) {                // Y is the major axis
        if(y1 > y2) {
            std::swap(y1, y2);
            std::swap(x1, x2);
        }

        int error = dy >> 1;
        int addx  = (x2 < x1) ? -1 : 1;

        /* The error term is the distance between the actual exact
           line and the pixel we set. It directly maps to an alpha
           value. Since our alpha runs from 0..255, the mapping is
           alpha=255*error/dy. To avoid dividing in a loop, we
           represent alpha as its integer part, and the fractional
           part as numerator of a fraction with base dy, and increment
           that step by step. */

        // Initial 'alpha = 255 * error / dy'. Can probably be simplified.
        uint8       alpha       = 255 * error / dy;
        int         alpha_fract = 255 * error % dy;

        // alpha increments by 255*dx/dy each iteration.
        uint8       alpha_inc       = 255 * dx / dy;
        int         alpha_fract_inc = 255 * dx % dy;

        while(y1 <= y2) {
            // draw pixel
            can.drawPixel(x1,      y1, parm.color, ~alpha);
            can.drawPixel(x1+addx, y1, parm.color,  alpha);

            // increment y, and error term (fractional X position)
            y1++;
            error += dx;

            // increment alpha
            alpha += alpha_inc;
            alpha_fract += alpha_fract_inc;
            if (alpha_fract >= dy) {
                alpha_fract -= dy;
                ++alpha;
            }

            if(error >= dy) {
                /* decrementing error by dy means decrementing alpha
                   by 255*dy/dy, i.e. incrementing by one */
                error -= dy;
                ++alpha;
                x1 += addx;
            }
        }
    } else {                // X is the major axis
        if(x1 > x2) {
            std::swap(y1, y2);
            std::swap(x1, x2);
        }

        int error = dx >> 1;
        int addy  = (y2 < y1) ? -1 : 1;

        uint8       alpha       = 255 * error / dx;
        int         alpha_fract = 255 * error % dx;

        uint8       alpha_inc       = 255 * dy / dx;
        int         alpha_fract_inc = 255 * dy % dx;
        
        while(x1 <= x2) {
            can.drawPixel(x1, y1,      parm.color, ~alpha);
            can.drawPixel(x1, y1+addy, parm.color,  alpha);
            x1++;
            error += dy;
            alpha += alpha_inc;
            alpha_fract += alpha_fract_inc;
            if (alpha_fract >= dx) {
                alpha_fract -= dx;
                ++alpha;
            }
            if(error >= dx) {
                error -= dx;
                ++alpha;
                y1 += addy;
            }
        }
    }
}

/** Circle drawing, with anti-aliasing. Draws a circle of radius r
    around x0,y0. We draw pixels closer than r+1 from center. */
void
drawCircleAA(GfxCanvas& can, GfxParameters& parm, int x0, int y0, int r)
{
    register int w = 0;
    register int k = 1;
    register int z = 0;
    register int y2 = 0;
    register int xk = 2*r-1;
    register int x = r;

    /* Given x, we need to compute y = sqrt(r**2 - x**2), or
       y**2 = r**2 - x**2. Starting with x=r, y starts at 0.

       Define z = w**2, and z >= y**2.
       
       We can incrementally compute exact values for y**2 = y2 using
       the identity that y**2 = \sum_{i=0..y} 2y-1 (bottom of outer loop).

       To find the smallest z such that z >= y**2, we compute the z and w
       values using the same identity (inner loop).

       This generates an y value for all x \in [0,r]. Since we're only
       interested in one octant, we filter out cases where w<x. */
    while (x >= 0) {
        while (z < y2) {
            z += k;
            ++w;
            k += 2;
        }
        if (w >= x) {
            /* The actual circle y2 lies between z and next z, i.e.
               z and z+k. Look where we are. This is not mathematically
               clean (should have some roots and squares here), but
               looks well in practice.

               FIXME: should get rid of this division instruction
               someday */
            uint8 alpha = (z - y2) * 255 / k;

            can.drawPixel(x0 - x, y0 - w, parm.color, ~alpha);
            can.drawPixel(x0 - x, y0 - w+1, parm.color, alpha);

            can.drawPixel(x0 + x, y0 - w, parm.color, ~alpha);
            can.drawPixel(x0 + x, y0 - w+1, parm.color, alpha);

            can.drawPixel(x0 - w, y0 - x, parm.color, ~alpha);
            can.drawPixel(x0 - w+1, y0 - x, parm.color, alpha);

            can.drawPixel(x0 + w, y0 - x, parm.color, ~alpha);
            can.drawPixel(x0 + w-1, y0 - x, parm.color, alpha);
            
            can.drawPixel(x0 - w, y0 + x, parm.color, ~alpha);
            can.drawPixel(x0 - w+1, y0 + x, parm.color, alpha);

            can.drawPixel(x0 + w, y0 + x, parm.color, ~alpha);
            can.drawPixel(x0 + w-1, y0 + x, parm.color, alpha);

            can.drawPixel(x0 - x, y0 + w, parm.color, ~alpha);
            can.drawPixel(x0 - x, y0 + w-1, parm.color, alpha);
            
            can.drawPixel(x0 + x, y0 + w, parm.color, ~alpha);
            can.drawPixel(x0 + x, y0 + w-1, parm.color, alpha);
        }
        y2 += xk;
        xk -= 2;
        --x;
    }
}

// /** Circle drawing, with anti-aliasing.  */
// void
// drawFilledCircleAA(GfxCanvas& can, GfxParameters& parm, int x0, int y0, int r)
// {
//     register int w = 0;
//     register int k = 1;
//     register int z = 0;
//     register int y2 = 0;
//     register int xk = 2*r-1;
//     register int x = r;

//     while (x >= 0) {
//         while (z < y2) {
//             z += k;
//             ++w;
//             k += 2;
//         }
//         if (w >= x) {
//             uint8 alpha = (z - y2) * 255 / k;

//             can.drawPixel(x0 - x, y0 - w, parm.color, ~alpha);
// //             can.drawPixel(x0 - x, y0 - w+1, parm.color, alpha);

//              can.drawPixel(x0 + x, y0 - w, parm.color, ~alpha);
// //             can.drawPixel(x0 + x, y0 - w+1, parm.color, alpha);

//              can.drawPixel(x0 - w, y0 - x, parm.color, ~alpha);
//              can.drawPixel(x0 + w, y0 - x, parm.color, ~alpha);
//              can.drawHLine(x0 - w+1, y0 - x, x0 + w, parm.color, 255, 255);
            
//              can.drawPixel(x0 - w, y0 + x, parm.color, ~alpha);
//              can.drawPixel(x0 + w, y0 + x, parm.color, ~alpha);
//              can.drawHLine(x0 - w+1, y0 + x, x0 + w, parm.color, 255, 255);

//              can.drawPixel(x0 - x, y0 + w, parm.color, ~alpha);
// //             can.drawPixel(x0 - x, y0 + w-1, parm.color, alpha);
            
//              can.drawPixel(x0 + x, y0 + w, parm.color, ~alpha);
// //             can.drawPixel(x0 + x, y0 + w-1, parm.color, alpha);
//         }
//         y2 += xk;
//         xk -= 2;
//         --x;
//     }
// }


static void
test(GfxCanvas& screen, GfxParameters& parm, int x1, int y1, int x2, int y2)
{
    drawLineAA(screen, parm, x1, y1, x2, y2);
    x1 += 320;
    x2 += 320;
    drawLine(screen, parm, x1, y1, x2, y2);
}

static void
testC(GfxCanvas& screen, GfxParameters& parm, int x1, int y1, int r)
{
    drawCircleAA(screen, parm, x1, y1, r);
    x1 += 320;
    drawCircle(screen, parm, x1, y1, r);
}

// static void
// testF(GfxCanvas& screen, GfxParameters& parm, int x1, int y1, int r)
// {
//     drawFilledCircleAA(screen, parm, x1, y1, r);
//     x1 += 320;
//     drawFilledCircle(screen, parm, x1, y1, r);
// }

int main()
{
    SDL_Surface* sfc = initGraphics(0,0);
    initPalette(sfc);
    GfxSurface screen(sfc, true);

    GfxParameters parm;
    parm.color = standard_colors[COLOR_RED];
    screen.lock();

    drawSolidBar(screen, GfxRect(screen), standard_colors[COLOR_WHITE]);

    for (int i = 0; i < 10; ++i) {
        drawSolidBar(screen, GfxRect(20*i, 0, 10, 100), standard_colors[COLOR_YELLOW]);
        drawSolidBar(screen, GfxRect(320+20*i, 0, 10, 100), standard_colors[COLOR_YELLOW]);
    }

#if 0
    /* a single line */
    test(screen, parm, 100, 1, 102, 40);
    test(screen, parm, 102, 41, 104, 80);
#endif
#if 0
    /* some random lines */
    for (int i = 0; i < 100; ++i) {
        int x1 = rand() % 300, y1 = rand() % 400, x2 = rand() % 300, y2 = rand() % 400;
        test(screen, parm, x1, y1, x2, y2);
    }
#endif
#if 0
    /* a moire circle */
    for (int i = 0; i < 360; ++i) {
        test(screen, parm, 150, 150, 150+int(150*sin(3.141595/180*i)+.5), 150+int(150*cos(3.141595/180*i)+.5));
    }
#endif
#if 0
    /* a circle outline (not a good algorithm, as many points are drawn twice) */
    for (int i = 0; i < 360; ++i) {
        test(screen, parm,
             150+int(150*sin(3.141595/180*i)+.5), 150+int(150*cos(3.141595/180*i)+.5),
             150+int(150*sin(3.141595/180*(i+1))+.5), 150+int(150*cos(3.141595/180*(i+1))+.5));
    }
#endif
#if 0
    /* some simple circles */
    testC(screen, parm, 100, 100, 50);
    testC(screen, parm, 8, 10, 5);
    testC(screen, parm, 20, 10, 4);
    testC(screen, parm, 30, 10, 3);
#endif
#if 1
    /* random circles */
    for (int i = 0; i < 100; ++i)
        testC(screen, parm, rand() % 100 + 100, rand() % 300, rand() % 80 + 20);
#endif
#if 0
    /* concentric circles without gaps; gives nice moire patterns in both cases */
    for (int i = 0; i < 100; ++i)
        testC(screen, parm, 105, 105, i);
#endif
#if 0
    for (int i = 0; i < 100; ++i)
        //testC(screen, parm, 10+i, 10+i, i);
        testC(screen, parm, 160-i/3, 160-i/3, i);
#endif
#if 0
    /* concentric circles with gap, gives nice circles with AA, moire without */
    for (int i = 0; i < 100; i += 2)
        testC(screen, parm, 105, 105, i);
#endif
// #if 1
//     /* some simple filled circles */
//     testF(screen, parm, 100, 100, 50);
//     testF(screen, parm, 8, 10, 5);
//     testF(screen, parm, 20, 10, 4);
//     testF(screen, parm, 30, 10, 3);
// #endif
    screen.unlock();
    screen.update(GfxRect(screen));

    getchar();
}
