/**
  *  \file ui/ui.h
  *  \brief Simple User Interface Kernel
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_UI_H
#define PCC_V2_UI_UI_H

#include <vector>
#include "gfx/gfx.h"
#include "ui/event.h"
#include "cpluslib/misc.h"
#include "cpluslib/string.h"
#include "cpluslib/smartptr.h"
#include "cpluslib/sigslot.h"
#include "cpluslib/holder.h"

class UIWidgetContainer;
class UILayoutManager;
class UISkin;

class UIWidget : public Object {
    UIWidget(const UIWidget&);
    UIWidget& operator=(const UIWidget&);
 protected:
    GfxRect            extent;  /**< Position and size of widget. Absolute coordinates. */
    UIWidgetContainer* owner;   /**< Owner of this widget. */
    UIWidget*          next;
    UIWidget*          prev;
    UISkin*            skin;
    
    int       status;           /**< Status flags. These change during lifetime. */
    int       options;          /**< Options. */
    int       id;               /**< Identifier. */

 public:
    struct LayoutInfo {
        /** Flags for layout (li_XXX). */
        int      flags;
        /** Minimum size. Layout manager should make sure widget does
            not get smaller than this. */
        GfxPoint min_size;
        /** Preferred size. When asked to compute the optimum size (pack()),
            all widgets come out with this size if possible. */
        GfxPoint pref_size;
        LayoutInfo();
    };
    enum {
        /** Widget does not want to be layouted. Layout managers will
            ignore it. */
        li_NoLayout = 1,
        /** Widget is flexible. It might be enlarged when there is
            more space than needed. */
        li_FlexibleH = 2,
        li_FlexibleV = 4,
        li_Flexible  = 6
    };
    enum {
        st_Exposed     = 1,     ///< Component is visible.
        st_Focused     = 2,     ///< Component has physical input focus.
        st_Selected    = 4,     ///< Component has logical input focus.
        st_Disabled    = 8,     ///< Component cannot be selected.
        st_IdleNotify  = 16,    ///< Component wants to receive `Idle' events. This flag only works for leaf widgets, not for UIWidgetContainer descendants.
        st_Modal       = 32,    ///< Component is modal.
        st_WantRedraw  = 64,    ///< Component wishes to be redrawn.
        st_ChildRedraw = 128,   ///< Component has a child that wishes to be redrawn.

        st_Last        = 128
    };
    Ptr<Object> layout_arg;     /** Layout argument. */

    /* Birth & Death */
    UIWidget() throw();
    UIWidget(int id) throw();
    virtual      ~UIWidget();

    /* Look & Feel */
    void          drawWidget(bool immediate);
    void          drawWidget();
    virtual void  draw(GfxCanvas& canvas) = 0;
    virtual void  drawDeferred(GfxCanvas& canvas);
    virtual bool  handleEvent(const UIEvent& event, bool second_pass);
    virtual bool  handleCommand(const UICommand& cmd, UIWidget& sender);
    virtual bool  isInside(int x, int y);

    virtual void  setSkin(UISkin& skin, bool recursive = false);
    UISkin&       getSkin() { return *skin; }
    bool          hasSkin() { return skin; }

    /* Status Management */
    int          getState() const { return status; }
    bool         hasState(int what) const { return (status & what) != 0; }
    void         setState(int astat, bool enable);
    virtual void stateChanged(int bits, bool enabled);

    void         setFocus();
    void         select();

    /* Administrivia */
    void                  setOwner(UIWidgetContainer* g) { owner = g; }
    UIWidgetContainer*    getOwner() const { return owner; }
    UIWidget*             getNext() const { return next; }
    UIWidget*             getPrev() const { return prev; }
    void                  setLinks(UIWidget* prev, UIWidget* next);

    int                   getId() const { return id; }
    virtual UIWidget*     getWidgetById(int id);
    virtual UIWidget*     getFocusedWidget();

    const GfxRect&        getExtent() const { return extent; }

    /* Size Changes */
    virtual void          getLayoutInfo(LayoutInfo&) = 0;
    virtual void          pack();
    virtual void          doLayout();

    void                  setSize(int w, int h);
    void                  moveTo(int x, int y);

    virtual void          setExtent(GfxRect newpos, bool by_layout = false);

    /* Prefix Argument */
    static void           setPrefixArg(int);
    static int            consumePrefixArg();

    /* Main loops */
    virtual int           run();
    virtual void          stop(int arg);

    string_t              toString();
};

class UIFixedWidget : public UIWidget {
 public:
    UIFixedWidget(int aid) throw() : UIWidget(aid) { }
    UIFixedWidget() throw() : UIWidget() { }
    ~UIFixedWidget();

    void getLayoutInfo(LayoutInfo& info);
};

class UIWidgetContainer : public UIWidget {
 protected:
    UIWidget* first;
    UIWidget* last;
    UIWidget* selected;
    UIWidget* focused;

 public:
    UIWidgetContainer(int aid = 0) throw();
    ~UIWidgetContainer();

    /* Look & Feel */
    virtual GfxCanvas&  getCanvasFor(UIWidget* child, ObjectHolder& h) = 0;
    virtual void        drawBelow(UIWidget* child, GfxCanvas& can) = 0;
    virtual void        drawDeferred(GfxCanvas& can);

    virtual GfxRect     getClientRectFor(GfxRect ext);
    virtual GfxRect     getExtentFor(GfxRect cli);

    virtual void        removeWidget(UIWidget& w) = 0;

    UIWidget*           getWidgetById(int id);
    UIWidget*           getFocusedWidget();
    virtual void        selectWidget(UIWidget* w);
    void                setFocus(UIWidget* w);
    void                setFocus() { UIWidget::setFocus(); }
    virtual void        handleChildResize(UIWidget* child, GfxRect oldsize) = 0;
    void                stateChanged(int astat, bool enable);
    void                setIdleNotify(bool enable);
    string_t            toString();

    virtual void add(UIWidget& widget) = 0;

    UIWidget* getFirstWidget() const
        { return first; }
    UIWidget* getLastWidget() const
        { return last; }
    UIWidget* getFocusedDescendant() const
        { return focused; }

    void setStateOf(int id, int astat, bool enable);
    void setSkin(UISkin& skin, bool recursive = false);

    Signal1<int>    sig_focus_changed;
};

class UIGroup : public UIWidgetContainer {
    UILayoutManager* layout;
 public:
    UIGroup(int aid = 0) throw();
    UIGroup(UILayoutManager& layout, int aid = 0) throw();
    ~UIGroup();

    void setLayout(UILayoutManager& alayout);
    UILayoutManager* getLayout() const;

    virtual void drawInterior(GfxCanvas& canvas);
    void draw(GfxCanvas& canvas);
    void drawBelow(UIWidget* child, GfxCanvas& can);

    bool handleEvent(const UIEvent& event, bool second_pass);
    GfxCanvas& getCanvasFor(UIWidget* child, ObjectHolder& h);
    void handleChildResize(UIWidget* child, GfxRect oldsize);
    void add(const GfxRect& apos, UIWidget& widget);
    void add(UIWidget& widget);
    void removeWidget(UIWidget& w);

    void setExtent(GfxRect r, bool by_layout = false);

    void getLayoutInfo(LayoutInfo&);
    void pack();
    void doLayout();
};

/// Spacer.
class UISpacer : public UIWidget {
 public:
    UISpacer() throw();
    void draw(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo&);
};

#endif
