/**
  *  \file ui/widget-button.cc
  *  \brief Push-button-like Widgets
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets.h"
#include "ui/event.h"
#include "gfx/gfxcompl.h"
#include "gfx/gfx.h"
#include "gfx/font.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "cpluslib/assert.h"

/**************************** UIAbstractButton ***************************/

/** \class UIAbstractButton
    \brief Standard Button

    This class provides the event handling mechanisms of a standard
    clickable button (checkbox, radiobutton, ...). The button can be
    have an associated hot-key if so desired.

    Buttons are triggered by a keystroke, mouse click and possibly a
    command.

    Derived classes must override draw().

    You can attach own handlers to sig_fired. */

/** Construct new key button. The button will post a UIKey event.
    \param key key that triggers this button. */
UIAbstractButton::UIAbstractButton(UIKey key)
    : UIWidget(),
      key(key)
{
    setState(bf_Key, true);
}

/** Construct new button.
    \param id widget id, used for emitting commands etc.
    \param key key that triggers this button */
UIAbstractButton::UIAbstractButton(int id, UIKey key)
    : UIWidget(id),
      key(key)
{ }

UIAbstractButton::~UIAbstractButton()
{ }

UIAbstractButton&
UIAbstractButton::addCommand()
{
    sig_fired.add(this, &UIButton::postCommand);
    return *this;
}

UIAbstractButton&
UIAbstractButton::addStop()
{
    sig_fired.add(this, &UIButton::callStop);
    return *this;
}

UIAbstractButton&
UIAbstractButton::addKey()
{
    setState(bf_Key, true);
    return *this;
}

/** Change button state. When the button loses logical focus, it must
    be un-pressed. Certain events also need redraw. */
void
UIAbstractButton::stateChanged(int astate, bool enable)
{
    UIWidget::stateChanged(astate, enable);

    if (astate & st_Disabled)
        status &= ~(bf_Pressed | st_Selected);

    if ((status & (bf_Pressed | st_Selected)) == bf_Pressed)
        status &= ~bf_Pressed;

    if (astate & (bf_Pressed | st_Selected | bf_Highlight | st_Disabled | bf_PressOverride))
        drawWidget(false);
}

/** Event handler. Processes mouse and keyboard events, and calls
    fire() when needed. */
bool
UIAbstractButton::handleEvent(const UIEvent& e, bool second_pass)
{
    switch(e.type) {
     case ev_Key:
        if (!hasState(st_Disabled) && !hasState(bf_Key) && e.key.code == key) {
            select();
            fire();
            return true;
        }
        break;
     case ev_Mouse:
        if (extent.contains(e.mouse.x, e.mouse.y) && !hasState(st_Disabled)) {
            /* mouse is in widget. */
            select();
            if (e.mouse.b) {
                /* mouse pressed in button */
                setState(bf_Pressed, true);
            } else if (hasState(bf_Pressed)) {
                /* mouse released in button */
                setState(bf_Pressed, false);
                if (hasState(bf_Key)) {
                    UIEvent ne;
                    ne.type         = ev_Key;
                    ne.key.sdl_type = SDL_KEYDOWN;
                    ne.key.arg      = 0;// FIXME
                    ne.key.code = key | UIEvent::simplMod(uint32(SDL_GetModState()) << 16);
                    ne.key.unicode  = 0;// FIXME?
                    putEvent(ne);
                }
                fire();
            }
            return true;
        } else {
            /* automatically lose selection when mouse is out. */
            if (hasState(st_Selected))
                owner->selectWidget(0);

            /* let others handle the event, too. */
            return false;
        }
        break;
     default:
        break;
    }
    return UIWidget::handleEvent(e, second_pass);
}

bool
UIAbstractButton::handleCommand(const UICommand& cmd, UIWidget& sender)
{
    if (cmd.code == cm_INT_QueryReferencePoint) {
        // FIXME: do we need this? This is the default setting.
        ASSERT(cmd.infop);
        TReferencePointArg& ra = *static_cast<TReferencePointArg*>(cmd.infop);
        ra.x      = getExtent().x;
        ra.y      = getExtent().getBottomY();
        return true;
    } else if (cmd.code == cm_INT_Push) {
        setState(bf_PressOverride, true);
    } else if (cmd.code == cm_INT_Unpush) {
        setState(bf_PressOverride, false);
    }
    return UIWidget::handleCommand(cmd, sender);
}

void
UIAbstractButton::postCommand(int prefix)
{
    handleCommand(UICommand(getId(), prefix), *this);
}

void
UIAbstractButton::callStop(int)
{
    stop(id);
}

/** Reaction on button click. */
void
UIAbstractButton::fire()
{
    sig_fired.raise(consumePrefixArg());
}

/******************************** UIButton *******************************/

/** \class UIButton
    \brief Push button

    This class provides the look of a standard push-button. */

UIButton::~UIButton()
{ }

void
UIButton::drawButton(GfxCanvas& can, GfxRect extent,
                     const string_t& name, int flags)
{
    // FIXME: use actual PCC I look. Maybe skinnable?
    const int fnr = (flags & bf_SmallSize) ? FONT_SMALL : FONT_TITLE;
    int delta = 0;
    if (flags & (bf_Pressed | bf_PressOverride)) {
        drawFrameDown(can, extent);
        ++delta;
    } else {
        drawFrameUp(can, extent);
    }

    int color = (flags & bf_Highlight
                 ? COLOR_GRAYSCALE + 12
                 : COLOR_GRAYSCALE + 9);
    if (flags & st_Selected)
        color++;

    can.drawBar(GfxRect(extent.x + 1, extent.y + 1,
                        extent.w - 2, extent.h - 2),
                standard_colors[color],
                GFX_TRANSPARENT,
                SOLID_FILL,
                GFX_ALPHA_OPAQUE);

    GfxParameters parm;
    if (flags & st_Disabled)
        parm.color = standard_colors[COLOR_BLACK];
    else
        parm.color = standard_colors[COLOR_SHIELD + 5];

    int x = extent.x + extent.w/2 - fonts[fnr]->getTextWidth(name)/2 + delta;
    int y = extent.y + extent.h/2 - fonts[fnr]->getTextHeight(name)/2 + delta;
    fonts[fnr]->outText(can, parm, x, y, name);
}

/** Draw the button. */
void
UIButton::draw(GfxCanvas& can)
{
    drawButton(can, extent, name, getState());
    can.update(extent);
}

void
UIButton::getLayoutInfo(LayoutInfo& info)
{
    if (hasState(bf_SmallSize)) {
        info.flags = 0;
        info.min_size = info.pref_size
            = GfxPoint(name.length() == 1 ? 15 : 8 + fonts[FONT_SMALL]->getTextWidth(name), 15);
    } else {
        info.flags = 0;
        info.min_size = info.pref_size
            = GfxPoint(name.length() == 1 ? 25 : 15 + fonts[FONT_TITLE]->getTextWidth(name), 25);
    }
}
