/**
  *  \file ui/window.cc
  *  \brief Top-Level Windows and Root Window
  *
  *  Implementation of the Top-Level Window widget and a root group.
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/window.h"
#include "gfx/multiclip.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "ui/util.h"
#include "ui/event.h"
#include "gfx/clip.h"
#include "cpluslib/assert.h"
#include "ui/widgets.h"
#include "gfx/update.h"

/**
  *  \class UIRoot
  *  \brief Root Group
  *
  *  This implements a ``root group'': a group which does not need
  *  a owner group.
  *
  *  This particular incarantion supports:
  *  - overlapping windows
  *  - painting on a GfxCanvas (which it expects to own completely)
  *  - nested invocations of the event loop
  */

/** Create Root Group on Canvas. The canvas should be a GfxSurface
    or something similar. */
UIRoot::UIRoot(Ptr<GfxCanvas> ascreen)
    : UIWidgetContainer(),
      multi_clip(*ascreen),
      screen(ascreen)
{
    setSkin(UISkin::getDefaultSkin());
    setExtent(GfxRect(*ascreen), true);
    setState(st_Exposed | st_Focused | st_Selected, true);
}

GfxCanvas&
UIRoot::getCanvasFor(UIWidget* child, ObjectHolder& h)
{
    sync();

    if (child == first || !child) {
        return *screen;
    }

    GfxMultiClip& mc = h.add(new GfxMultiClip(*screen));
    mc.add(GfxRect(*screen));
    for (UIWidget* p = first; p && p != child; p = p->getNext())
        mc.remove(p->getExtent());

    return mc;
}

/** Draw Background. Child classes can override this to do something
    more interesting. \default fills the background with COLOR_BLACK. */
void
UIRoot::drawBackground(GfxCanvas& can)
{
    can.drawBar(getExtent(), standard_colors[COLOR_BLACK], GFX_TRANSPARENT,
                SOLID_FILL, GFX_ALPHA_OPAQUE);
    can.update(getExtent());
}

void
UIRoot::draw(GfxCanvas& can)
{
    GfxMultiClip mc(can);
    mc.add(GfxRect(*screen));

    for (UIWidget* p = first; p; p = p->getNext()) {
        p->draw(mc);
        mc.remove(p->getExtent());
    }
    if (mc.isVisible(getExtent()))
        drawBackground(mc);
}

void
UIRoot::drawBelow(UIWidget* child, GfxCanvas& can)
{
    sync();

    GfxMultiClip mc(can);
    mc.add(GfxRect(*screen));
    for (UIWidget* p = child->getNext(); p; p = p->getNext()) {
        p->draw(mc);
        mc.remove(p->getExtent());
    }
    if (mc.isVisible(getExtent()))
        drawBackground(mc);
}

/** Draw frames around current window's widgets. This is used to
    visualize layout management.
    \param can Screen canvas
    \param p   Root widget

    Widgets are color-coded as follows:
    - red: UIGroup (all descendants shown)
    - white: UICardGroup (only visible descendant shown)
    - green, slashed: UISpacer
    - yellow: regular widgets */
static void
doFrames(GfxCanvas& can, UIWidget* p)
{
    if (!p || !p->hasState(UIWidget::st_Exposed))
        return;

    GfxParameters parm;
    if (UICardGroup* c = dynamic_cast<UICardGroup*>(p)) {
        doFrames(can, c->getFocusedDescendant());
        parm.color = standard_colors[COLOR_WHITE];
    } else if (UIWidgetContainer* g = dynamic_cast<UIWidgetContainer*>(p)) {
        for (UIWidget* pp = g->getFirstWidget(); pp != 0; pp = pp->getNext())
            doFrames(can, pp);
        parm.color = standard_colors[COLOR_RED];
    } else if (dynamic_cast<UISpacer*>(p)) {
        parm.fill_pattern = LTSLASH_FILL;
        parm.color = standard_colors[COLOR_GREEN];
        drawBar(can, parm, p->getExtent());
    } else {
        parm.color = standard_colors[COLOR_YELLOW];
    }
    drawRectangle(can, parm, p->getExtent());
}

/** Event Handling. This does the focus policy:
    - keyboard and command events go to topmost window only;
    - mouse events go to all windows that can see it;
    - rest (timer, broadcast) goes to everyone. */
bool
UIRoot::handleEvent(const UIEvent& e, bool second_pass)
{
    /* events we do ourselves */
    if (e.type == ev_Key) {
        if (e.key.code == ss_Ctrl + ss_Shift + 's') {
            doScreenshot();
            return true;
        }
        if (e.key.code == ss_Ctrl + ss_Shift + 'q') {
            putKey(vk_Quit);
            return true;
        }
        if (e.key.code == ss_Ctrl + ss_Shift + 'f') {
            doFrames(*screen, first);
            screen->update(getExtent());
            return true;
        }
    }

    /* propagate stuff */
    if (e.type == ev_Key) {
        /* top widget only if modal, propagate otherwise. */
        for (UIWidget* p = first; p; p = p->getNext()) {
            if (p->handleEvent(e, second_pass))
                return true;
            if (p->hasState(st_Modal))
                return false;
        }
        return false;
    }
    if (e.type == ev_Mouse) {
        /* widgets which see it only */
        int mx = e.mouse.x;
        int my = e.mouse.y;
        for (UIWidget* p = first; p; p = p->getNext()) {
            if (p->handleEvent(e, second_pass))
                return true;
            if (p->isInside(mx, my) || p->hasState(st_Modal))
                return false;
        }
        return false;
    }
    /* other widgets: broadcast */
    for (UIWidget* p = first; p; p = p->getNext())
        if (p->handleEvent(e, second_pass))
            return true;

    /* UIWidget::handleEvent, actually */
    return UIWidgetContainer::handleEvent(e, second_pass);
}

int
UIRoot::run()
{
    stop_requested = false;
    while (!stop_requested) {
        UIEvent e;
        if (!haveQueuedEvent())
            doDeferredRedraws();
        getEvent(e, hasState(st_IdleNotify));
        if (!handleEvent(e, false)) {
            if (!handleEvent(e, true)) {
                /* If it is an unhandled user event, clear the prefix
                   argument. User events are keypresses and mouse button
                   releases. Since prefix argument entry is terminated
                   by either a keypress, or a pressed mouse button, it
                   suffices to test for b==0. This either means a button
                   release as required, or we don't have an argument. */
                if (e.type == ev_Key || (e.type == ev_Mouse && e.mouse.b == 0))
                    consumePrefixArg();
            }
        }
    }
    stop_requested = false;  /* in case handleEvent() calls run again */
    return stop_arg;
}

/** Run widget. The widget usually is a modal window (but needn't be).
    This function call combines the add(), run() and close() calls in
    one call. */
int
UIRoot::runModal(UIWidget& w)
{
    add(w);
    int result = run();
    close(w);
    return result;
}

void
UIRoot::stop(int arg)
{
    stop_arg       = arg;
    stop_requested = true;
}

/** Process Timer Event. Post timer event to topmost window, if needed.
    Call this occasionally while doing expensive computations. This
    permits display of `idle' indicators.

    \note that this is different from standard timer events which are
    posted to all windows. This one goes to the topmost window only
    for performance reasons. Wo don't want to degrade the performance
    of the above-mentioned heavy processing by stupid event handling. */
void
UIRoot::postTimerEvent()
{
    static long old_timer = 0;
    long this_timer = getTimer();

    if (this_timer == old_timer)
        return;

    old_timer = this_timer;

    UIEvent e;
    e.type = ev_Tick;
    if (first)
        first->handleEvent(e, false) || first->handleEvent(e, true);
}

/** Center Widget on Screen. */
void
UIRoot::centerWidget(UIWidget& w)
{
    // FIXME: could be member of UIWidgetContainer
    GfxRect ext = w.getExtent();
    ext.centerWithin(getExtent());
    w.setExtent(ext);
}

/** Move widget to a "logical" position. The position is given by two
    alignment values, as for text alignment.
    \param w      corpus delicti
    \param x_pos  horizontal position: 0=left, 1=center, 2=right
    \param y_pos  vertical position: 0=top, 1=center, 2=right
    \param offset distance to edge. At offset 10, the window will stay
                  10 pixels from screen border. */
void
UIRoot::moveWidgetToEdge(UIWidget& w, int x_pos, int y_pos, int offset)
{
    // FIXME: could be member of UIWidgetContainer
    GfxRect ext = w.getExtent();
    ext.moveToEdge(getExtent(), x_pos, y_pos, offset);
    w.setExtent(ext);
}

/** Handle Child Resize.
    \todo Right now, this is almost a copy of UIGroup::handleChildResize(). */
void
UIRoot::handleChildResize(UIWidget* child, GfxRect oldsize)
{
    sync();
    if (child->getExtent().contains(oldsize)) {
        /* child completely covers old position. */
        child->drawWidget(true);
    } else {
        /* child has moved. Complete redraw in case something previously
           obscured (i.e. background) now becomes visible. */
        GfxMultiClip mc(*screen);
        mc.add(child->getExtent());
        mc.add(oldsize);

        // FIXME: lock/unlock?
        GfxUpdateOptimizer updater(mc);
        draw(updater);
        updater.doUpdates();
    }
}

/** Add Window. Adds the window as the top-most (current) window.
    It need not be a UIWindow. */
void
UIRoot::add(UIWidget& w)
{
    multi_clip.remove(w.getExtent());
    sync();
    if (hasState(st_Exposed) && screen->getBitsPerPixel() > 8) {
#if 0
        // FIXME: software rot. These effects crash when a listbox is
        // opened (probably anything without its own skin)
#if 1
        GfxRect size = w.getExtent();
        Ptr<GfxPixmap> pix = new GfxPixmap(size.w, size.h, *screen);
        Ptr<GfxCanvas> sfc = new GfxOffset(-size.x, -size.y, *pix);
        GfxUpdateOptimizer updater(*sfc);
        w.draw(updater);
        updater.doUpdates();

        UIPollableTimer timer(30);
        static const int offsets[] = {
            0, 8, 4, 12, 2, 10, 6, 14, 3, 11, 7, 15, 5, 13, 9, 1
        };
        for (int i = 0; i < 16; ++i) {
            for (int y = offsets[i]; y < size.h; y += 16)
                screen->blitSurface(size.x, size.y, *pix,
                                    GfxRect(0, y, size.w, 1));
            screen->update(size);
            timer.waitTime(i+1);
        }
#else
        /* funky effect stolen from "Master of Antares" by EAD */
        UIPollableTimer timer(30);
        GfxRect size = w.getExtent();
#if 1
        for (int i = 1; i < 10; i += 2) {
            screen->drawBar(size, standard_colors[COLOR_WHITE], GFX_TRANSPARENT,
                            SOLID_FILL, i*i*2);
            screen->update(size);
            timer.waitTime(i/2+1);
        }
        screen->drawBar(size, standard_colors[COLOR_WHITE], GFX_TRANSPARENT,
                        SOLID_FILL, GFX_ALPHA_OPAQUE);
#endif
        Ptr<GfxPixmap> pix = new GfxPixmap(size.w, size.h, *screen);
        Ptr<GfxCanvas> sfc = new GfxOffset(-size.x, -size.y, *pix);
        GfxUpdateOptimizer updater(*sfc);
        w.draw(updater);
        updater.doUpdates();

        timer.setTime(0);
        for (int i = 1; i < 10; i += 2) {
            pix->setBlitAlpha(i*i*2);
            screen->blitSurface(size.x, size.y, *pix, GfxRect(0, 0, size.w, size.h));
            screen->update(size);
            timer.waitTime(i/2+1);
        }
#endif
#endif
    }

    /* build window stack */
    if (first)
        first->setLinks(&w, first->getNext());
    w.setLinks(0, first);
    w.setOwner(this);
    first = &w;
    if (!last)
        last = &w;
    setFocus(0);                // remove focus, so add will give it to our window
    UIWidgetContainer::add(w);
    selectWidget(&w);
}

/** Remove the specified widget. */
void
UIRoot::removeWidget(UIWidget& w)
{
    if (&w == first) {
        /* closing topmost window */
        multi_clip.add(w.getExtent());
        UIWidgetContainer::removeWidget(w);
    } else {
        /* closing "middle" window */
        sync();
        UIWidget* p = first;
        while (p && p->getNext() != &w)
            p = p->getNext();
        ASSERT(p);              // if this fails, w is not a child of us

        UIWidgetContainer::removeWidget(w);

        /* FIXME: this can be optimized to not redraw windows before w */
        /* FIXME: lock/unlock? */
        GfxClipRectFilter cr(w.getExtent(), *screen);
        GfxUpdateOptimizer updater(cr);
        draw(updater);
        updater.doUpdates();
    }
}

/** Redraw everything. Redraws the whole screen. */
void
UIRoot::redraw()
{
    sync();
    // FIXME: lock/unlock?
    GfxUpdateOptimizer updater(*screen);
    draw(updater);
    updater.doUpdates();
}

void
UIRoot::doDeferredRedraws()
{
    sync();
    // FIXME: lock/unlock?
    GfxUpdateOptimizer updater(*screen);
    drawDeferred(updater);
    updater.doUpdates();
}

void
UIRoot::sync()
{
    if (multi_clip.empty())
        return;
    // FIXME: lock/unlock?
    GfxUpdateOptimizer updater(multi_clip);
    draw(updater);
    updater.doUpdates();
    multi_clip.clear();
}

void
UIRoot::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_NoLayout;
}

/******************************** UIWindow *******************************/

/**
  *  \class UIWindow
  *  \brief Standard Window
  *
  *  This is a standard framed window.
  */

Ptr<GfxPixmap> UIWindow::frame_blue;

UIWindow::UIWindow(const GfxRect& size, string_t aname, int afs, int abs,
                   int aborder)
    : UIGroup(0),
      skin(*this),
      name(aname),
      frame_style(afs),
      bg_style(abs),
      border_space(aborder)
{
    setExtent(size, true);
    if (!frame_blue.ptr())
        frame_blue = requestPixmap("winblue");
    setSkin(skin);
    setState(st_Modal, true);
}

UIWindow::UIWindow(string_t aname, int afs, int abs, int aborder)
    : UIGroup(0),
      skin(*this),
      name(aname),
      frame_style(afs),
      bg_style(abs),
      border_space(aborder)
{
    if (!frame_blue.ptr())
        frame_blue = requestPixmap("winblue");
    setSkin(skin);
    setState(st_Modal, true);
}

/** Draw a window. Available as static function to be callable from
    outside (widgets that look like windows but aren't, like the VCR
    screen) */
void
UIWindow::draw(GfxCanvas& canvas, const GfxRect& extent,
               int bg_style, string_t name, bool bg)
{
    GfxParameters parm;
    parm.color   = standard_colors[COLOR_WHITE];
    parm.just_x  = 1;

    drawFrameUp(canvas, extent);
    drawTiledArea(canvas, GfxRect(extent.x + 1, extent.y + 1,
                                  extent.w - 2, 22),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    drawTiledArea(canvas, GfxRect(extent.x + 1, extent.y + 23,
                                  2, extent.h - 26),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    drawTiledArea(canvas, GfxRect(extent.x + extent.w - 3, extent.y + 23,
                                  2, extent.h - 26),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    drawTiledArea(canvas, GfxRect(extent.x + 1, extent.y + extent.h - 3,
                                  extent.w - 2, 2),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);

    drawFrameDown(canvas, GfxRect(extent.x + 3, extent.y + 23,
                                  extent.w - 6, extent.h - 26));
    if (bg)
        // synchronize this with drawInterior
        canvas.drawBar(GfxRect(extent.x + 4, extent.y + 24,
                               extent.w - 8, extent.h - 28),
                       standard_colors[bg_style],
                       GFX_TRANSPARENT,
                       SOLID_FILL,
                       GFX_ALPHA_OPAQUE);

    parm.font = fonts[FONT_TITLE];
    outTextF(canvas, parm, extent.x + extent.w/2, extent.y + 2, extent.w, name);
}

void
UIWindow::draw(GfxCanvas& canvas)
{
    UIWindow::draw(canvas, getExtent(), bg_style, name, false);
    canvas.update(getExtent());
    UIGroup::draw(canvas);
}

void
UIWindow::drawInterior(GfxCanvas& canvas)
{
    GfxRect r(extent.x + 4, extent.y + 24,
              extent.w - 8, extent.h - 28);
    skin.drawBackground(canvas, r);
    canvas.update(r);
}

/** Event Handling. Standard event handling, owner does focus policy. */
bool
UIWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    return UIGroup::handleEvent(e, second_pass);
}

GfxRect
UIWindow::getClientRectFor(GfxRect ext)
{
    ext.w -= 8 + 2*border_space;
    ext.h -= 28 + 2*border_space;
    ext.x += 4 + border_space;
    ext.y += 24 + border_space;

    return ext;
}

GfxRect
UIWindow::getExtentFor(GfxRect cli)
{
    cli.w += 8 + 2*border_space;
    cli.h += 28 + 2*border_space;
    cli.x -= 4 + border_space;
    cli.y -= 24 + border_space;

    int min_w = fonts[FONT_TITLE]->getTextWidth(name) + 6;
    if (cli.w < min_w)
        cli.w = min_w;

    return cli;
}

const UIColorSet&
UIWindowSkin::getColors() const
{
    // FIXME: do we need that?
//     if (color == tc_Background && window.bg_style < 256) {
//         parm.color = standard_colors[window.bg_style];
//     }
    if (window.bg_style == COLOR_GRAY)
        return GRAY_COLOR_SET;
    else if (window.bg_style == COLOR_DARK)
        return DARK_COLOR_SET;
    else
        return GRAY_COLOR_SET;  // fallback, will probably be changed
}

void
UIWindowSkin::outText(GfxCanvas& can,  int f, int x, int y, string_t text)
{
    parm.font = fonts[f];
    parm.bgcolor = isSolid() ? standard_colors[window.bg_style] : GFX_TRANSPARENT;
    ::outText(can, parm, x, y, text);
}

void
UIWindowSkin::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    parm.font = fonts[f];
    parm.bgcolor = isSolid() ? standard_colors[window.bg_style] : GFX_TRANSPARENT;
    ::outTextF(can, parm, x, y, lim, text);
}

void
UIWindowSkin::drawBackground(GfxCanvas& can, GfxRect r)
{
    // synchronize this with static draw()
    can.drawBar(r, standard_colors[window.bg_style], GFX_TRANSPARENT,
                SOLID_FILL, GFX_ALPHA_OPAQUE);
}
